# future
from __future__ import annotations

# stdlib
from typing import Any
from typing import Dict
from typing import Iterable
from typing import List
from typing import Optional
from typing import TYPE_CHECKING
from typing import Tuple
from typing import Union

# third party
import numpy as np
from scipy.ndimage.interpolation import rotate

# relative
from ..common.serde.serializable import serializable
from .broadcastable import is_broadcastable
from .passthrough import is_acceptable_simple_type  # type: ignore
from .smpc.utils import get_shape

if TYPE_CHECKING:
    # relative
    from .autodp.phi_tensor import PhiTensor


@serializable(recursive_serde=True)
class lazyrepeatarray:
    """
    When data is repeated along one or more dimensions, store it using lazyrepeatarray
    so that you can save on RAM and CPU when computing with it. Think like the opposite
    of np.broadcast, repeated values along an axis are collapsed but the .shape
    attribute of the higher dimensional projection is retained for operations.
    """

    __attr_allowlist__ = ["data", "shape"]

    def __init__(self, data: np.ndarray, shape: Tuple[int, ...]) -> None:
        """
        data: the raw data values without repeats
        shape: the shape of 'data' if repeats were included
        """

        # NOTE: all additional arguments are assumed to be broadcast if dims are shorter
        # than that of data. Example: if data.shape == (2,3,4) and
        # min_vals.shape == (2,3), then it's assumed that the full min_vals.shape is
        # actually (2,3,4) where the last dim is simply copied.
        # Example2: if data.shape == (2,3,4) and min_vals.shape == (2,1,4), then the
        # middle dimension is supposed to be copied to be min_vals.shape == (2,3,4)
        # if necessary. This is just to keep the memory footprint (and computation)
        # as small as possible.

        if isinstance(data, (bool, int, float)):
            data = np.array(data)

        # verify broadcasting works on shapes
        if -1 not in shape:
            np.broadcast_shapes(data.shape, shape)

        self.data = data
        self.shape = shape
        if isinstance(shape, Iterable):
            for val in shape:
                if val < 0:
                    raise ValueError(f"Invalid shape: {shape}")

    def __getitem__(self, item: Union[str, int, slice]) -> lazyrepeatarray:
        if self.data.shape == self.shape:
            output = self.data[item]
            return lazyrepeatarray(data=output, shape=output.shape)
        elif self.data.size == 1:
            test_arr = np.ones(self.shape)[
                item
            ]  # TODO: Is there a better way to determine output shape?
            return lazyrepeatarray(data=self.data, shape=test_arr.shape)
        else:
            raise NotImplementedError

    def __add__(self, other: Any) -> lazyrepeatarray:
        """
        THIS MIGHT LOOK LIKE COPY-PASTED CODE!
        Don't touch it. It's going to get more complicated.
        """
        if is_acceptable_simple_type(other):
            return self.__class__(data=self.data + other, shape=self.shape)

        if not is_broadcastable(self.shape, other.shape):
            raise Exception(
                f"Cannot broadcast arrays with shapes: {self.shape} & {other.shape}"
            )

        if self.data.shape == other.data.shape:
            return self.__class__(data=self.data + other.data, shape=self.shape)
        else:
            print("Lazy Repeat adding with mismatched shapes")
            return self.__class__(data=self.data + other.data, shape=self.shape)

    def __sub__(self, other: Any) -> lazyrepeatarray:
        """
        THIS MIGHT LOOK LIKE COPY-PASTED CODE!
        Don't touch it. It's going to get more complicated.
        """
        if is_acceptable_simple_type(other):
            res = self.data - other
            return self.__class__(data=res, shape=res.shape)

        if not is_broadcastable(self.shape, other.shape):
            raise Exception(
                f"Cannot broadcast arrays with shapes: {self.shape} & {other.shape}"
            )

        if self.data.shape == other.data.shape:
            return self.__class__(data=self.data - other.data, shape=self.shape)
        else:
            print("Lazy Repeat adding with mismatched shapes")
            return self.__class__(data=self.data - other.data, shape=self.shape)

    def __mul__(self, other: Any) -> lazyrepeatarray:
        """
        THIS MIGHT LOOK LIKE COPY-PASTED CODE!
        Don't touch it. It's going to get more complicated.
        """
        if is_acceptable_simple_type(other):
            return self.__class__(data=self.data * other, shape=self.shape)

        if not is_broadcastable(self.shape, other.shape):
            raise Exception(
                "Cannot broadcast arrays with shapes for LazyRepeatArray Multiplication:"
                + f" {self.shape} & {other.shape}"
            )
        else:
            return self.__class__(data=self.data * other.data, shape=self.shape)

    def __matmul__(self, other: Any) -> lazyrepeatarray:
        """
        THIS MIGHT LOOK LIKE COPY-PASTED CODE!
        Don't touch it. It's going to get more complicated.
        """
        if is_acceptable_simple_type(other):
            new_shape = get_shape("__matmul__", self.shape, other.shape)

            if self.data.size == 1:
                return self.__class__(
                    data=np.matmul(np.ones(self.shape), other * self.data),
                    shape=new_shape,
                )
            return self.__class__(data=self.data.__matmul__(other), shape=new_shape)

        if self.shape[-1] != other.shape[-2]:
            raise Exception(
                f"cannot matrix multiply tensors with different shapes: {self.shape} and {other.shape}"
            )

        result = self.to_numpy() @ other.to_numpy()
        return self.__class__(data=result, shape=result.shape)

        # raise Exception("not sure how to do this yet")

    def zeros_like(self, *args: Tuple[Any, ...], **kwargs: Any) -> lazyrepeatarray:
        res = np.array(np.zeros_like(self.to_numpy(), *args, **kwargs))
        return lazyrepeatarray(data=res, shape=res.shape)

    def __rtruediv__(self, other: Any) -> lazyrepeatarray:
        res = (1 / self.data) * other
        return lazyrepeatarray(data=res, shape=self.shape)

    def __rmatmul__(self, other: Any) -> lazyrepeatarray:
        """
        THIS MIGHT LOOK LIKE COPY-PASTED CODE!
        Don't touch it. It's going to get more complicated.
        """
        if is_acceptable_simple_type(other):
            new_shape = get_shape("__matmul__", other.shape, self.shape)

            if other.size == 1:
                return self.__class__(
                    data=np.matmul(np.ones(other.shape), other * self.data),
                    shape=new_shape,
                )
            return self.__class__(
                data=self.to_numpy().__rmatmul__(other), shape=new_shape
            )

        if other.shape[-1] != self.shape[0]:
            raise Exception(
                "cannot matrix multiply tensors with different shapes: {self.shape} and {other.shape}"
            )

        result = self.to_numpy().__rmatmul__(other.to_numpy())
        return self.__class__(data=result, shape=result.shape)

    def __pow__(self, exponent: int) -> lazyrepeatarray:
        if exponent == 2:
            return self * self
        raise Exception("not sure how to do this yet")

    def pad(self, pad_width: int, mode: str = "reflect") -> lazyrepeatarray:
        if mode == "reflect":
            new_shape = tuple([i + pad_width * 2 for i in self.shape])
            if self.data.shape == self.shape:
                return lazyrepeatarray(
                    data=np.pad(self.data, pad_width=pad_width, mode="reflect"),
                    shape=new_shape,
                )
            elif self.data.size == 1:
                return lazyrepeatarray(data=self.data, shape=new_shape)
            else:
                raise NotImplementedError
        else:
            raise NotImplementedError

    def horizontal_flip(self) -> lazyrepeatarray:
        if self.data.shape == self.shape:
            return lazyrepeatarray(data=np.fliplr(self.data), shape=self.shape)
        elif self.data.size == 1:
            return lazyrepeatarray(data=self.data, shape=self.shape)
        else:
            raise NotImplementedError

    def vertical_flip(self) -> lazyrepeatarray:
        if self.data.shape == self.shape:
            return lazyrepeatarray(data=np.flipud(self.data), shape=self.shape)
        elif self.data.size == 1:
            return lazyrepeatarray(data=self.data, shape=self.shape)
        else:
            raise NotImplementedError

    def rotate(self, angle: int) -> lazyrepeatarray:
        if self.data.shape == self.shape:
            return lazyrepeatarray(data=rotate(self.data, angle), shape=self.shape)
        elif self.data.size == 1:
            # TODO: This is almost certainly incorrect
            return lazyrepeatarray(data=self.data, shape=self.shape)
        else:
            raise NotImplementedError

    def reshape(self, target_shape: Tuple) -> lazyrepeatarray:
        if self.data.shape == self.shape:
            return lazyrepeatarray(
                data=self.data.reshape(target_shape), shape=target_shape
            )
        elif self.data.size == 1:
            return lazyrepeatarray(data=self.data, shape=target_shape)
        else:
            if not np.broadcast_shapes(self.data.shape, target_shape):
                raise NotImplementedError(
                    f"data= {self.data.shape}, shape: {self.shape}"
                )
            else:
                return lazyrepeatarray(data=self.data, shape=target_shape)

    def copy(self, order: Optional[str] = "K") -> lazyrepeatarray:
        return self.__class__(data=self.data.copy(order=order), shape=self.shape)

    @property
    def size(self) -> int:
        return np.prod(self.shape)

    def sum(self, *args: Tuple[Any, ...], **kwargs: Any) -> lazyrepeatarray:
        res = np.array(self.to_numpy().sum(*args, **kwargs))
        return lazyrepeatarray(data=res, shape=res.shape)

    def ones_like(self, *args: Tuple[Any, ...], **kwargs: Any) -> lazyrepeatarray:
        res = np.array(np.ones_like(self.to_numpy(), *args, **kwargs))
        return lazyrepeatarray(data=res, shape=res.shape)

    def __eq__(self, other: Any) -> lazyrepeatarray:  # type: ignore
        if isinstance(other, lazyrepeatarray):
            if self.shape == other.shape:
                return lazyrepeatarray(data=self.data == other.data, shape=self.shape)
            else:
                result = (self.to_numpy() == other.to_numpy()).all()
                return lazyrepeatarray(data=np.array([result]), shape=result.shape)
        if isinstance(other, np.ndarray):
            try:
                _ = np.broadcast_shapes(self.shape, other.shape)
                result = (self.to_numpy() == other).all()
                return lazyrepeatarray(data=np.array([result]), shape=other.shape)
            except Exception as e:
                print(
                    "Failed to compare lazyrepeatarray with "
                    + f"{self.shape} == {other.shape} to numpy by broadcasting. {e}"
                )
                raise e

        return self == other

    def __le__(self, other: Any) -> lazyrepeatarray:  # type: ignore
        if isinstance(other, lazyrepeatarray):
            if self.shape == other.shape:
                return lazyrepeatarray(data=self.data <= other.data, shape=self.shape)
            else:
                result = (self.to_numpy() <= other.to_numpy()).all()
                return lazyrepeatarray(data=np.array([result]), shape=result.shape)
        if isinstance(other, np.ndarray):
            try:
                _ = np.broadcast_shapes(self.shape, other.shape)
                result = (self.to_numpy() <= other).all()
                return lazyrepeatarray(data=np.array([result]), shape=other.shape)
            except Exception as e:
                print(
                    "Failed to compare lazyrepeatarray with "
                    + f"{self.shape} == {other.shape} to numpy by broadcasting. {e}"
                )
                raise e

        return self <= other

    def concatenate(
        self, other: lazyrepeatarray, *args: List[Any], **kwargs: Dict[str, Any]
    ) -> lazyrepeatarray:
        if not isinstance(other, lazyrepeatarray):
            raise NotImplementedError

        dummy_res = np.concatenate(
            (np.empty(self.shape), np.empty(other.shape)), *args, **kwargs
        )
        return lazyrepeatarray(data=self.data, shape=dummy_res.shape)

    @property
    def dtype(self) -> np.dtype:
        return self.data.dtype

    def astype(self, np_type: np.dtype) -> lazyrepeatarray:
        return self.__class__(self.data.astype(np_type), self.shape)

    def to_numpy(self) -> np.ndarray:
        return np.broadcast_to(self.data, self.shape)

    def __repr__(self) -> str:
        return f"<lazyrepeatarray data: {self.data} -> shape: {self.shape}>"

    def __bool__(self) -> bool:
        return self.data.__bool__()

    def all(self) -> bool:
        return self.data.all()

    def any(self) -> bool:
        return self.data.any()

    def transpose(self, *args: List[Any], **kwargs: Dict[str, Any]) -> lazyrepeatarray:
        dummy_res = self.to_numpy().transpose(*args, **kwargs)
        return lazyrepeatarray(
            data=self.data.transpose(*args, **kwargs), shape=dummy_res.shape
        )


# As the min and max values calculation is the same regardless of the tensor type,
# We centralize this method as baseline for calculation for min/max values
def compute_min_max(
    x_min_vals: lazyrepeatarray,
    x_max_vals: lazyrepeatarray,
    other: Union[PhiTensor, int, float, np.ndarray],
    op_str: str,
) -> Tuple[lazyrepeatarray, lazyrepeatarray]:
    min_vals: lazyrepeatarray
    max_vals: lazyrepeatarray

    if op_str in ["__add__", "__matmul__", "__rmatmul__"]:
        if is_acceptable_simple_type(other):
            min_vals = getattr(x_min_vals, op_str)(other)
            max_vals = getattr(x_max_vals, op_str)(other)
        elif hasattr(other, "min_vals") and hasattr(other, "max_vals"):
            min_vals = getattr(x_min_vals, op_str)(other.min_vals)  # type: ignore
            max_vals = getattr(x_max_vals, op_str)(other.max_vals)  # type: ignore
        else:
            raise ValueError(
                f"Not supported type for lazy repeat array computation: {type(other)}"
            )

    elif op_str in ["__sub__", "__mul__"]:
        if is_acceptable_simple_type(other):
            min_vals = getattr(x_min_vals, op_str)(other)
            max_vals = getattr(x_max_vals, op_str)(other)
        elif hasattr(other, "min_vals") and hasattr(other, "max_vals"):
            min_min = getattr(x_min_vals.data, op_str)(other.min_vals.data)  # type: ignore
            min_max = getattr(x_min_vals.data, op_str)(other.max_vals.data)  # type: ignore
            max_min = getattr(x_max_vals.data, op_str)(other.min_vals.data)  # type: ignore
            max_max = getattr(x_max_vals.data, op_str)(other.max_vals.data)  # type: ignore
            _min_vals = np.minimum.reduce([min_min, min_max, max_min, max_max])
            _max_vals = np.maximum.reduce([min_min, min_max, max_min, max_max])
            min_vals = x_min_vals.copy()
            min_vals.data = _min_vals
            max_vals = x_max_vals.copy()
            max_vals.data = _max_vals
        else:
            raise ValueError(
                f"Not supported type for lazy repeat array computation: {type(other)}"
            )

    elif op_str in ["__gt__", "__lt__", "__le__", "__ge__", "__eq__", "__ne__"]:
        min_vals = x_min_vals * 0
        max_vals = (x_max_vals * 0) + 1
    elif op_str == "sum":
        min_vals = lazyrepeatarray(data=np.array(x_min_vals.sum(axis=None)), shape=())
        max_vals = lazyrepeatarray(data=np.array(x_max_vals.sum(axis=None)), shape=())
    else:
        raise ValueError(f"Invaid Operation for LazyRepeatArray: {op_str}")

    return (min_vals, max_vals)



# stdlib
from typing import Any
from typing import Dict

# relative
from ...abstract.node import AbstractNodeClient
from ...enums import AssociationRequestResponses
from ...enums import RequestAPIFields
from ...enums import ResponseObjectEnum
from ...exceptions import PyGridClientException
from ..node_service.association_request.association_request_messages import (
    DeleteAssociationRequestMessage,
)
from ..node_service.association_request.association_request_messages import (
    GetAssociationRequestMessage,
)
from ..node_service.association_request.association_request_messages import (
    GetAssociationRequestsMessage,
)
from ..node_service.association_request.association_request_messages import (
    RespondAssociationRequestMessage,
)
from ..node_service.association_request.association_request_messages import (
    SendAssociationRequestMessage,
)
from ..node_service.success_resp_message import ErrorResponseMessage
from .request_api import RequestAPI


class AssociationRequestAPI(RequestAPI):
    def __init__(self, client: AbstractNodeClient):
        super().__init__(
            client=client,
            create_msg=SendAssociationRequestMessage,
            get_msg=GetAssociationRequestMessage,
            get_all_msg=GetAssociationRequestsMessage,
            delete_msg=DeleteAssociationRequestMessage,
            response_key=ResponseObjectEnum.ASSOCIATION_REQUEST,
        )

    def create(self, **kwargs: Any) -> None:
        retry = kwargs.pop("retry")
        response = self.perform_api_request(  # type: ignore
            syft_msg=self._create_message, content=kwargs
        )

        if isinstance(response, ErrorResponseMessage):
            if retry > 0:
                print(response.resp_msg, " Retrying", retry, " ...")
                kwargs["retry"] = retry - 1
                self.create(**kwargs)
            else:
                raise Exception(response.resp_msg)

    def update(self, **kwargs: Any) -> Dict[Any, Any]:  # type: ignore
        raise PyGridClientException(
            "You can not update an association request, try to send another one instead."
        )

    def get(self, **kwargs: Any) -> Any:
        association_table = self.perform_api_request(
            syft_msg=self._get_message, content=kwargs
        )

        content = getattr(
            association_table, "content", getattr(association_table, "metadata", None)
        )
        if content is None:
            raise Exception(f"{type(self)} has no content or metadata field")

        obj_data = dict(content)
        obj_data[RequestAPIFields.SOURCE] = association_table.source
        obj_data[RequestAPIFields.TARGET] = association_table.target

        return self.to_obj(obj_data)

    def __getitem__(self, key: int) -> Any:
        return self.get(association_id=key)

    def __delitem__(self, key: int) -> None:
        self.delete(association_id=key)

    def to_obj(self, result: Dict[Any, Any]) -> Any:
        _association_obj = super().to_obj(result)

        _content = {
            RequestAPIFields.SOURCE: result[RequestAPIFields.SOURCE],
            RequestAPIFields.TARGET: result[RequestAPIFields.TARGET],
        }

        def _accept() -> Dict[str, str]:
            _content[RequestAPIFields.RESPONSE] = AssociationRequestResponses.ACCEPT
            return self.perform_api_request(
                syft_msg=RespondAssociationRequestMessage, content=_content
            )

        def _deny() -> Dict[str, str]:
            _content[RequestAPIFields.RESPONSE] = AssociationRequestResponses.DENY
            return self.perform_api_request(
                syft_msg=RespondAssociationRequestMessage, content=_content
            )

        _association_obj.accept = _accept
        _association_obj.deny = _deny

        return _association_obj



# stdlib
from typing import List
from typing import Optional

# relative
from .uid import UID


class AbstractStorableObject:

    data: object
    id: UID
    search_permissions: dict

    @property
    def icon(self) -> str:
        return "🗂️"

    @property
    def pprint(self) -> str:
        output = f"{self.icon} ({self.class_name})"
        return output

    @property
    def class_name(self) -> str:
        return str(self.__class__.__name__)

    @property
    def object_type(self) -> str:
        raise NotImplementedError

    @property
    def tags(self) -> Optional[List[str]]:
        raise NotImplementedError

    @property
    def description(self) -> Optional[str]:
        raise NotImplementedError

    @property
    def is_proxy(self) -> bool:
        raise NotImplementedError



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: proto/grid/messages/dataset_messages.proto
"""Generated protocol buffer code."""
# third party
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


# syft absolute
from syft.proto.core.common import (
    common_object_pb2 as proto_dot_core_dot_common_dot_common__object__pb2,
)
from syft.proto.core.io import address_pb2 as proto_dot_core_dot_io_dot_address__pb2

DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n*proto/grid/messages/dataset_messages.proto\x12\x12syft.grid.messages\x1a%proto/core/common/common_object.proto\x1a\x1bproto/core/io/address.proto"\xac\x02\n\x14\x43reateDatasetMessage\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12&\n\x07\x61\x64\x64ress\x18\x02 \x01(\x0b\x32\x15.syft.core.io.Address\x12\x0f\n\x07\x64\x61taset\x18\x03 \x01(\x0c\x12H\n\x08metadata\x18\x04 \x03(\x0b\x32\x36.syft.grid.messages.CreateDatasetMessage.MetadataEntry\x12\'\n\x08reply_to\x18\x05 \x01(\x0b\x32\x15.syft.core.io.Address\x12\x10\n\x08platform\x18\x06 \x01(\t\x1a/\n\rMetadataEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01"\x9f\x01\n\x11GetDatasetMessage\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12&\n\x07\x61\x64\x64ress\x18\x02 \x01(\x0b\x32\x15.syft.core.io.Address\x12\x12\n\ndataset_id\x18\x03 \x01(\x03\x12\'\n\x08reply_to\x18\x04 \x01(\x0b\x32\x15.syft.core.io.Address"\xdc\x01\n\x12GetDatasetResponse\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12\x46\n\x08metadata\x18\x02 \x03(\x0b\x32\x34.syft.grid.messages.GetDatasetResponse.MetadataEntry\x12&\n\x07\x61\x64\x64ress\x18\x03 \x01(\x0b\x32\x15.syft.core.io.Address\x1a/\n\rMetadataEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01"\x8c\x01\n\x12GetDatasetsMessage\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12&\n\x07\x61\x64\x64ress\x18\x02 \x01(\x0b\x32\x15.syft.core.io.Address\x12\'\n\x08reply_to\x18\x03 \x01(\x0b\x32\x15.syft.core.io.Address"\xd7\x02\n\x13GetDatasetsResponse\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12M\n\tmetadatas\x18\x02 \x03(\x0b\x32:.syft.grid.messages.GetDatasetsResponse.metadata_container\x12&\n\x07\x61\x64\x64ress\x18\x03 \x01(\x0b\x32\x15.syft.core.io.Address\x1a\xa1\x01\n\x12metadata_container\x12Z\n\x08metadata\x18\x01 \x03(\x0b\x32H.syft.grid.messages.GetDatasetsResponse.metadata_container.MetadataEntry\x1a/\n\rMetadataEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\x0c:\x02\x38\x01"\x9d\x02\n\x14UpdateDatasetMessage\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12&\n\x07\x61\x64\x64ress\x18\x02 \x01(\x0b\x32\x15.syft.core.io.Address\x12\x12\n\ndataset_id\x18\x03 \x01(\x03\x12H\n\x08metadata\x18\x04 \x03(\x0b\x32\x36.syft.grid.messages.UpdateDatasetMessage.MetadataEntry\x12\'\n\x08reply_to\x18\x05 \x01(\x0b\x32\x15.syft.core.io.Address\x1a/\n\rMetadataEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01"\xb9\x01\n\x14\x44\x65leteDatasetMessage\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12&\n\x07\x61\x64\x64ress\x18\x02 \x01(\x0b\x32\x15.syft.core.io.Address\x12\x12\n\ndataset_id\x18\x03 \x01(\t\x12\x15\n\rbin_object_id\x18\x04 \x01(\t\x12\'\n\x08reply_to\x18\x05 \x01(\x0b\x32\x15.syft.core.io.Addressb\x06proto3'
)

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(
    DESCRIPTOR, "proto.grid.messages.dataset_messages_pb2", globals()
)
if _descriptor._USE_C_DESCRIPTORS == False:

    DESCRIPTOR._options = None
    _CREATEDATASETMESSAGE_METADATAENTRY._options = None
    _CREATEDATASETMESSAGE_METADATAENTRY._serialized_options = b"8\001"
    _GETDATASETRESPONSE_METADATAENTRY._options = None
    _GETDATASETRESPONSE_METADATAENTRY._serialized_options = b"8\001"
    _GETDATASETSRESPONSE_METADATA_CONTAINER_METADATAENTRY._options = None
    _GETDATASETSRESPONSE_METADATA_CONTAINER_METADATAENTRY._serialized_options = b"8\001"
    _UPDATEDATASETMESSAGE_METADATAENTRY._options = None
    _UPDATEDATASETMESSAGE_METADATAENTRY._serialized_options = b"8\001"
    _CREATEDATASETMESSAGE._serialized_start = 135
    _CREATEDATASETMESSAGE._serialized_end = 435
    _CREATEDATASETMESSAGE_METADATAENTRY._serialized_start = 388
    _CREATEDATASETMESSAGE_METADATAENTRY._serialized_end = 435
    _GETDATASETMESSAGE._serialized_start = 438
    _GETDATASETMESSAGE._serialized_end = 597
    _GETDATASETRESPONSE._serialized_start = 600
    _GETDATASETRESPONSE._serialized_end = 820
    _GETDATASETRESPONSE_METADATAENTRY._serialized_start = 388
    _GETDATASETRESPONSE_METADATAENTRY._serialized_end = 435
    _GETDATASETSMESSAGE._serialized_start = 823
    _GETDATASETSMESSAGE._serialized_end = 963
    _GETDATASETSRESPONSE._serialized_start = 966
    _GETDATASETSRESPONSE._serialized_end = 1309
    _GETDATASETSRESPONSE_METADATA_CONTAINER._serialized_start = 1148
    _GETDATASETSRESPONSE_METADATA_CONTAINER._serialized_end = 1309
    _GETDATASETSRESPONSE_METADATA_CONTAINER_METADATAENTRY._serialized_start = 1262
    _GETDATASETSRESPONSE_METADATA_CONTAINER_METADATAENTRY._serialized_end = 1309
    _UPDATEDATASETMESSAGE._serialized_start = 1312
    _UPDATEDATASETMESSAGE._serialized_end = 1597
    _UPDATEDATASETMESSAGE_METADATAENTRY._serialized_start = 388
    _UPDATEDATASETMESSAGE_METADATAENTRY._serialized_end = 435
    _DELETEDATASETMESSAGE._serialized_start = 1600
    _DELETEDATASETMESSAGE._serialized_end = 1785
# @@protoc_insertion_point(module_scope)



# stdlib
from typing import Any
from typing import Optional
from typing import Tuple
from typing import Union

# third party
import numpy as np

# relative
from ....common.serde.serializable import serializable
from ...autodp.gamma_tensor import GammaTensor
from ...autodp.phi_tensor import PhiTensor
from .base import Layer


@serializable(recursive_serde=True)
class Flatten(Layer):
    __attr_allowlist__ = ("outdim", "last_input_shape", "out_shape", "input_shape")

    def __init__(self, outdim: int = 2) -> None:
        self.outdim = outdim
        if outdim < 1:
            raise ValueError("Dim must be >0, was %i", outdim)

        self.last_input_shape: Optional[Tuple[int, ...]] = None
        self.out_shape: Optional[Tuple[int, ...]] = None

    def connect_to(self, prev_layer: Layer) -> None:
        if len(prev_layer.out_shape) < 2:  # type: ignore
            raise ValueError(
                f"Input layer shape should have more than two dimensions:{prev_layer.out_shape}"  # type: ignore
            )

        to_flatten = np.prod(prev_layer.out_shape[self.outdim - 1 :])  # type: ignore # noqa: E203
        flattened_shape = prev_layer.out_shape[: self.outdim - 1] + (int(to_flatten),)  # type: ignore

        self.input_shape = prev_layer.out_shape  # type: ignore
        self.out_shape = flattened_shape

    def forward(
        self, input: Union[PhiTensor, GammaTensor], *args: Any, **kwargs: Any
    ) -> Union[PhiTensor, GammaTensor]:
        self.last_input_shape = input.shape

        to_flatten = np.prod(self.last_input_shape[self.outdim - 1 :])  # noqa: E203
        flattened_shape = input.shape[: self.outdim - 1] + (int(to_flatten),)
        return input.reshape(flattened_shape)

    def backward(
        self, pre_grad: Union[PhiTensor, GammaTensor], *args: Any, **kwargs: Any
    ) -> Union[PhiTensor, GammaTensor]:
        if self.last_input_shape is None:
            raise ValueError(
                "`self.last_input_shape` cannot be None. \
                Please check the shape of the last input."
            )
        return pre_grad.reshape(self.last_input_shape)



# third party
from sqlalchemy import Column
from sqlalchemy import JSON
from sqlalchemy import String

# relative
from . import Base


class JsonObject(Base):
    __tablename__ = "json_object"

    id = Column(String(), primary_key=True)
    binary = Column(JSON())

    def __str__(self) -> str:
        return f"<JsonObject id: {self.id}>"



# stdlib
from typing import Any
from typing import List

# relative
from ..manager import TensorChainManager
from .mpc_tensor import MPCTensor
from .utils import ispointer


class MPCTensorAncestor(TensorChainManager):
    def share(self, *parties: List[Any]) -> MPCTensor:
        # relative
        from .mpc_tensor import MPCTensor

        if ispointer(self.child):
            raise ValueError(
                "Cannot call share on a remote tensor. Use MPCTensor(remote_secret)"
            )

        return MPCTensor(secret=self.child, parties=list(parties))



# stdlib
import sys
from typing import Any
from typing import Dict
from typing import Iterator
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

# third party
from google.protobuf.reflection import GeneratedProtocolMessageType
from nacl.signing import SigningKey
from nacl.signing import VerifyKey
import pandas as pd

# syft absolute
import syft as sy

# relative
from ....grid import GridURL
from ....logger import critical
from ....logger import debug
from ....logger import error
from ....logger import info
from ....logger import traceback_and_raise
from ....proto.core.node.common.client_pb2 import Client as Client_PB
from ....proto.core.node.common.metadata_pb2 import Metadata as Metadata_PB
from ....util import get_fully_qualified_name
from ...common.message import EventualSyftMessageWithoutReply
from ...common.message import ImmediateSyftMessageWithReply
from ...common.message import ImmediateSyftMessageWithoutReply
from ...common.message import SignedEventualSyftMessageWithoutReply
from ...common.message import SignedImmediateSyftMessageWithReply
from ...common.message import SignedImmediateSyftMessageWithoutReply
from ...common.message import SignedMessage
from ...common.message import SyftMessage
from ...common.serde.serializable import serializable
from ...common.uid import UID
from ...io.location import Location
from ...io.location import SpecificLocation
from ...io.route import Route
from ...pointer.garbage_collection import GarbageCollection
from ...pointer.garbage_collection import gc_get_default_strategy
from ...pointer.pointer import Pointer
from ..abstract.node import AbstractNodeClient
from ..common.client_manager.node_networking_api import NodeNetworkingAPI
from .action.exception_action import ExceptionMessage
from .node_service.object_search.obj_search_service import ObjectSearchMessage


@serializable()
class Client(AbstractNodeClient):
    """Client is an incredibly powerful abstraction in Syft. We assume that,
    no matter where a client is, it can figure out how to communicate with
    the Node it is supposed to point to. If I send you a client I have
    with all of the metadata in it, you should have all the information
    you need to know to interact with a node (although you might not
    have permissions - clients should not store private keys)."""

    def __init__(
        self,
        name: Optional[str],
        routes: List[Route],
        network: Optional[Location] = None,
        domain: Optional[Location] = None,
        device: Optional[Location] = None,
        vm: Optional[Location] = None,
        signing_key: Optional[SigningKey] = None,
        verify_key: Optional[VerifyKey] = None,
        version: Optional[str] = None,
    ):
        name = f"{name}" if name is not None else None
        super().__init__(
            name=name, network=network, domain=domain, device=device, vm=vm
        )

        self.routes = routes
        self.default_route_index = 0

        gc_strategy_name = gc_get_default_strategy()
        self.gc = GarbageCollection(gc_strategy_name)

        # create a signing key if one isn't provided
        if signing_key is None:
            self.signing_key = SigningKey.generate()
        else:
            self.signing_key = signing_key

        # if verify key isn't provided, get verify key from signing key
        if verify_key is None:
            self.verify_key = self.signing_key.verify_key
        else:
            self.verify_key = verify_key

        self.install_supported_frameworks()

        self.store = StoreClient(client=self)
        self.networking = NodeNetworkingAPI(client=self)
        self.version = version

    def obj_exists(self, obj_id: UID) -> bool:
        raise NotImplementedError

    @property
    def icon(self) -> str:
        icon = "📡"
        sub = []
        if self.vm is not None:
            sub.append("🍰")
        if self.device is not None:
            sub.append("📱")
        if self.domain is not None:
            sub.append("🏰")
        if self.network is not None:
            sub.append("🔗")

        if len(sub) > 0:
            icon = f"{icon} ["
            for s in sub:
                icon += s
            icon += "]"
        return icon

    @staticmethod
    def deserialize_client_metadata_from_node(
        metadata: Metadata_PB,
    ) -> Tuple[SpecificLocation, str, UID]:
        # string of bytes
        meta = sy.deserialize(blob=metadata)
        return meta.node, meta.name, meta.id

    def install_supported_frameworks(self) -> None:
        self.lib_ast = sy.lib.create_lib_ast(client=self)

        # first time we want to register for future updates
        self.lib_ast.register_updates(self)

        if self.lib_ast is not None:
            for attr_name, attr in self.lib_ast.attrs.items():
                setattr(self, attr_name, attr)

        # shortcut syft.lib.python to just python
        if hasattr(self.lib_ast, "syft"):
            try:
                lib_attr = getattr(self.lib_ast.syft, "lib", None)

                if lib_attr is not None:
                    python_attr = getattr(lib_attr, "python", None)
                    self.python = python_attr
                    python_attr = getattr(lib_attr, "adp", None)
                    self.adp = python_attr

            except Exception as e:
                critical(f"Failed to set python attribute on client. {e}")

    def configure(self, **kwargs: Any) -> Any:
        # relative
        from .node_service.node_setup.node_setup_messages import UpdateSetupMessage

        if "daa_document" in kwargs.keys():
            kwargs["daa_document"] = open(kwargs["daa_document"], "rb").read()
        else:
            kwargs["daa_document"] = b""
        response = self._perform_grid_request(  # type: ignore
            grid_msg=UpdateSetupMessage, content=kwargs
        ).content
        info(response)

    @property
    def settings(self, **kwargs: Any) -> Dict[Any, Any]:  # type: ignore
        try:
            # relative
            from .node_service.node_setup.node_setup_messages import GetSetUpMessage

            return self._perform_grid_request(  # type: ignore
                grid_msg=GetSetUpMessage, content=kwargs
            ).content  # type : ignore
        except Exception:  # nosec
            # unable to fetch settings
            return {}

    def join_network(
        self,
        client: Optional[AbstractNodeClient] = None,
        host_or_ip: Optional[str] = None,
    ) -> None:
        # this asks for a VPN key so it must be on a public interface hence the
        # client or a public host_or_ip
        try:
            if client is None and host_or_ip is None:
                raise ValueError(
                    "join_network requires a Client object or host_or_ip string"
                )

            # we are leaving the client and entering the node in a container
            # any hostnames of localhost need to be converted to docker-host
            if client is not None:
                grid_url = client.routes[0].connection.base_url  # type: ignore
            else:
                grid_url = GridURL.from_url(str(host_or_ip))

            return self.vpn.join_network_vpn(grid_url=grid_url)  # type: ignore
        except Exception as e:
            msg = f"Failed to join network with {client} or {host_or_ip}. {e}"
            raise Exception(msg)

    @property
    def id(self) -> UID:
        """This client points to an node, this returns the id of that node."""
        traceback_and_raise(NotImplementedError)

    # TODO fix the msg type but currently tensor needs SyftMessage

    def send_immediate_msg_with_reply(
        self,
        msg: Union[
            SignedImmediateSyftMessageWithReply,
            ImmediateSyftMessageWithReply,
            Any,  # TEMPORARY until we switch everything to NodeRunnableMessage types.
        ],
        timeout: Optional[float] = None,
        return_signed: bool = False,
        route_index: int = 0,
    ) -> Union[SyftMessage, SignedMessage]:

        # relative
        from .node_service.simple.simple_messages import NodeRunnableMessageWithReply

        # TEMPORARY: if message is instance of NodeRunnableMessageWithReply then we need to wrap it in a SimpleMessage
        if isinstance(msg, NodeRunnableMessageWithReply):
            msg = msg.prepare(address=self.address, reply_to=self.address)

        route_index = route_index or self.default_route_index

        if isinstance(msg, ImmediateSyftMessageWithReply):
            output = (
                f"> {self.pprint} Signing {msg.pprint} with "
                + f"{self.key_emoji(key=self.signing_key.verify_key)}"
            )
            debug(output)
            msg = msg.sign(signing_key=self.signing_key)

        response = self.routes[route_index].send_immediate_msg_with_reply(
            msg=msg, timeout=timeout
        )
        if response.is_valid:
            # check if we have an ExceptionMessage to trigger a local exception
            # from a remote exception that we caused
            if isinstance(response.message, ExceptionMessage):
                exception_msg = response.message
                exception = exception_msg.exception_type(exception_msg.exception_msg)
                error(str(exception))
                traceback_and_raise(exception)
            else:
                if return_signed:
                    return response
                return response.message

        traceback_and_raise(
            Exception("Response was signed by a fake key or was corrupted in transit.")
        )

    # TODO fix the msg type but currently tensor needs SyftMessage

    def send_immediate_msg_without_reply(
        self,
        msg: Union[
            SignedImmediateSyftMessageWithoutReply, ImmediateSyftMessageWithoutReply
        ],
        route_index: int = 0,
        timeout: Optional[float] = None,
    ) -> None:
        route_index = route_index or self.default_route_index

        if isinstance(msg, ImmediateSyftMessageWithoutReply):
            output = (
                f"> {self.pprint} Signing {msg.pprint} with "
                + f"{self.key_emoji(key=self.signing_key.verify_key)}"
            )
            debug(output)
            msg = msg.sign(signing_key=self.signing_key)
        debug(f"> Sending {msg.pprint} {self.pprint} ➡️  {msg.address.pprint}")
        self.routes[route_index].send_immediate_msg_without_reply(
            msg=msg, timeout=timeout
        )

    def send_eventual_msg_without_reply(
        self,
        msg: EventualSyftMessageWithoutReply,
        route_index: int = 0,
        timeout: Optional[float] = None,
    ) -> None:
        route_index = route_index or self.default_route_index
        output = (
            f"> {self.pprint} Signing {msg.pprint} with "
            + f"{self.key_emoji(key=self.signing_key.verify_key)}"
        )
        debug(output)
        signed_msg: SignedEventualSyftMessageWithoutReply = msg.sign(
            signing_key=self.signing_key
        )

        self.routes[route_index].send_eventual_msg_without_reply(
            msg=signed_msg, timeout=timeout
        )

    def url_from_path(self, path: str) -> str:
        new_url = GridURL.from_url(url=path)
        client_url = self.routes[0].connection.base_url.copy()  # type: ignore
        new_url.protocol = client_url.protocol
        new_url.port = client_url.port
        new_url.host_or_ip = client_url.host_or_ip
        return new_url.url

    def __repr__(self) -> str:
        return f"<Client pointing to node with id:{self.id}>"

    def register_route(self, route: Route) -> None:
        self.routes.append(route)

    def set_default_route(self, route_index: int) -> None:
        self.default_route = route_index

    def _object2proto(self) -> Client_PB:
        client_pb = Client_PB(
            obj_type=get_fully_qualified_name(obj=self),
            id=sy.serialize(self.id),
            name=self.name,
            routes=[sy.serialize(route) for route in self.routes],
            network=self.network._object2proto() if self.network else None,
            domain=self.domain._object2proto() if self.domain else None,
            device=self.device._object2proto() if self.device else None,
            vm=self.vm._object2proto() if self.vm else None,
        )
        return client_pb

    @staticmethod
    def _proto2object(proto: Client_PB) -> "Client":
        module_parts = proto.obj_type.split(".")
        klass = module_parts.pop()
        obj_type = getattr(sys.modules[".".join(module_parts)], klass)

        obj = obj_type(
            name=proto.name,
            routes=[sy.deserialize(route) for route in proto.routes],
            network=sy.deserialize(proto.network)
            if proto.HasField("network")
            else None,
            domain=sy.deserialize(proto.domain) if proto.HasField("domain") else None,
            device=sy.deserialize(proto.device) if proto.HasField("device") else None,
            vm=sy.deserialize(proto.vm) if proto.HasField("vm") else None,
        )

        if type(obj) != obj_type:
            traceback_and_raise(
                TypeError(
                    f"Deserializing Client. Expected type {obj_type}. Got {type(obj)}"
                )
            )

        return obj

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        return Client_PB

    @property
    def keys(self) -> str:
        verify = (
            self.key_emoji(key=self.signing_key.verify_key)
            if self.signing_key is not None
            else "🚫"
        )
        keys = f"🔑 {verify}"

        return keys

    def __hash__(self) -> Any:
        return hash(self.id)


GET_OBJECT_TIMEOUT = 60  # seconds


class StoreClient:
    def __init__(self, client: Client) -> None:
        self.client = client

    @property
    def store(self) -> List[Pointer]:
        msg = ObjectSearchMessage(
            address=self.client.address, reply_to=self.client.address
        )

        results = getattr(
            self.client.send_immediate_msg_with_reply(
                msg=msg, timeout=GET_OBJECT_TIMEOUT
            ),
            "results",
            None,
        )
        if results is None:
            traceback_and_raise(ValueError("TODO"))

        # This is because of a current limitation in Pointer where we cannot
        # serialize a client object. TODO: Fix limitation in Pointer so that we don't need this.
        for result in results:
            result.gc_enabled = False
            result.client = self.client

        return results

    def __len__(self) -> int:
        """Return the number of items in the object store we're allowed to know about"""
        return len(self.store)

    def __iter__(self) -> Iterator[Any]:
        return self.store.__iter__()

    def __getitem__(self, key: Union[str, int, UID]) -> Pointer:
        return self.get(key=key)

    #
    # def __getitem__(self, key: Union[str, int, UID]) -> Pointer:
    #
    #     if isinstance(key, int):
    #         return self.store[key]
    #     elif isinstance(key, str):
    #         # PART 1: try using the key as an ID
    #         try:
    #             key = UID.from_string(key)
    #             return self[key]
    #         except ValueError:
    #
    #             # If there's no id of this key, then try matching on a tag
    #             matches = 0
    #             match_obj: Optional[Pointer] = None
    #
    #             for obj in self.store:
    #                 if key in obj.tags:
    #                     matches += 1
    #                     match_obj = obj
    #             if matches == 1 and match_obj is not None:
    #                 return match_obj
    #             else:  # matches > 1
    #                 traceback_and_raise(
    #                     KeyError("More than one item with tag:" + str(key))
    #                 )
    #                 raise KeyError("More than one item with tag:" + str(key))
    #
    #     elif isinstance(key, UID):
    #         msg = ObjectSearchMessage(
    #             address=self.client.address, reply_to=self.client.address, obj_id=key
    #         )
    #
    #         results = getattr(
    #             self.client.send_immediate_msg_with_reply(msg=msg), "results", None
    #         )
    #         if results is None:
    #             traceback_and_raise(ValueError("TODO"))
    #
    #         # This is because of a current limitation in Pointer where we cannot
    #         # serialize a client object. TODO: Fix limitation in Pointer so that we don't need this.
    #         for result in results:
    #             result.gc_enabled = False
    #             result.client = self.client
    #         if len(results) == 1:
    #             return results[0]
    #         return results
    #     else:
    #         traceback_and_raise(KeyError("Please pass in a string or int key"))

    def get(self, key: Union[str, int, UID]) -> Pointer:
        if isinstance(key, str):
            try:
                return self.get(UID.from_string(key))
            except IndexError:
                matches = 0
                match_obj: Optional[Pointer] = None

                for obj in self.store:
                    if key in obj.tags:
                        matches += 1
                        match_obj = obj
                if matches == 1 and match_obj is not None:
                    return match_obj
                elif matches > 1:
                    traceback_and_raise(
                        KeyError("More than one item with tag:" + str(key))
                    )
                else:
                    # If key does not math with any tags, we then try to match it with id string.
                    # But we only do this if len(key)>=5, because if key is too short, for example
                    # if key="a", there are chances of mismatch it with id string, and I don't
                    # think the user pass a key such short as part of id string.
                    str_key = str(key)
                    if len(str_key) >= 5:
                        for obj in self.store:
                            if str_key in str(obj.id_at_location.value).replace(
                                "-", ""
                            ):
                                return obj
                    else:
                        traceback_and_raise(
                            KeyError(
                                f"No such item found for tag: {key}, and we "
                                + "don't consider it as part of id string because its too short."
                            )
                        )

                traceback_and_raise(KeyError("No such item found for id:" + str(key)))
        if isinstance(key, int):
            return self.store[key]
        elif isinstance(key, UID):
            msg = ObjectSearchMessage(
                address=self.client.address, reply_to=self.client.address, obj_id=key
            )
            results = getattr(
                self.client.send_immediate_msg_with_reply(
                    msg=msg, timeout=GET_OBJECT_TIMEOUT
                ),
                "results",
                None,
            )

            if results is None:
                traceback_and_raise(ValueError("TODO"))

            # This is because of a current limitation in Pointer where we cannot
            # serialize a client object. TODO: Fix limitation in Pointer so that we don't need this.
            for result in results:
                result.gc_enabled = False
                result.client = self.client

            return results[0]
        else:
            traceback_and_raise(KeyError("Please pass in a string or int key"))

    def __repr__(self) -> str:
        return repr(self.store)

    @property
    def pandas(self) -> pd.DataFrame:
        obj_lines: List[Dict[str, Any]] = list()
        for obj in self.store:
            obj_lines.append(
                {
                    "ID": obj.id_at_location,
                    "Tags": obj.tags,
                    "Description": obj.description,
                    "object_type": obj.object_type,
                }
            )
        return pd.DataFrame(obj_lines)

    def _repr_html_(self) -> str:
        return self.pandas._repr_html_()



"""Crypto Primitives."""

# stdlib
from collections import defaultdict
from typing import Any
from typing import Callable
from typing import DefaultDict
from typing import Dict
from typing import List
from typing import Optional

# relative
from ...common import UID


class CryptoPrimitiveProvider:
    """A trusted third party should use this class to generate crypto primitives."""

    _func_providers: Dict[str, Callable] = {}
    _ops_list: DefaultDict[str, List] = defaultdict(list)

    def __init__(self) -> None:  # noqa
        raise ValueError("This class should not be initialized")

    @staticmethod
    def generate_primitives(
        op_str: str,
        parties: List[Any],
        g_kwargs: Optional[Dict[str, Any]] = None,
        p_kwargs: Optional[Dict[str, Any]] = None,
        nr_instances: int = 1,
        ring_size: int = 2**32,
    ) -> List[Any]:
        g_kwargs = g_kwargs if g_kwargs is not None else {}
        p_kwargs = p_kwargs if p_kwargs is not None else {}
        """Generate "op_str" primitives.

        Args:
            op_str (str): Operator.
            parties (List[Any]): Parties to generate primitives for.
            g_kwargs: Generate kwargs passed to the registered function.
            p_kwargs: Populate kwargs passed to the registered populate function.

        Returns:
            List[Any]: List of primitives.

        Raises:
            ValueError: If op_str is not registered.

        """
        if op_str not in CryptoPrimitiveProvider._func_providers:
            raise ValueError(f"{op_str} not registered")

        nr_parties = len(parties)

        generator = CryptoPrimitiveProvider._func_providers[op_str]
        primitives = generator(
            **g_kwargs,
            nr_instances=nr_instances,
            nr_parties=nr_parties,
            ring_size=ring_size,
        )

        if p_kwargs is not None:
            """Do not transfer the primitives if there is not specified a
            values for populate kwargs."""
            CryptoPrimitiveProvider._transfer_primitives_to_parties(
                op_str=op_str,
                primitives=primitives,
                parties=parties,
                p_kwargs=p_kwargs,
                ring_size=ring_size,
            )

        return primitives

    @staticmethod
    def _transfer_primitives_to_parties(
        op_str: str,
        primitives: List[Any],
        parties: List[Any],
        p_kwargs: Dict[str, Any],
        ring_size: int,
    ) -> None:
        # relative
        from ...node.common.action.beaver_primitive_action import BeaverPrimitiveAction

        if not isinstance(primitives, list):
            raise ValueError("Primitives should be a List")

        if len(primitives) != len(parties):
            raise ValueError(
                f"Primitives length {len(primitives)} != Parties length {len(parties)}"
            )

        for primitives_party, client in zip(primitives, parties):

            args: List[Any] = []
            kwargs = {
                "op_str": op_str,
                "primitives": primitives_party,
                **p_kwargs,
                "ring_size": str(ring_size),
            }
            path = "syft.core.tensor.smpc.share_tensor.populate_store"
            cmd = BeaverPrimitiveAction(
                path=path,
                args=args,
                kwargs=kwargs,
                id_at_location=UID(),
                address=client.address,
            )
            client.send_immediate_msg_without_reply(msg=cmd)

    @staticmethod
    def get_state() -> str:
        """Get the state of a CryptoProvider.

        Returns:
            str: CryptoProvider
        """
        res = f"Providers: {list(CryptoPrimitiveProvider._func_providers.keys())}\n"
        return res



# future
from __future__ import annotations

# stdlib
from functools import partial
import os
from pathlib import Path
import time
from typing import Any
from typing import Callable
from typing import List
from typing import Optional
from typing import TYPE_CHECKING
from typing import Tuple

# third party
from typing_extensions import Final

# relative
from ...logger import info

if TYPE_CHECKING:
    # stdlib
    from dataclasses import dataclass
else:
    from flax.struct import dataclass

# third party
import jax
from jax import numpy as jnp
from nacl.signing import VerifyKey
import numpy as np
from scipy.optimize import minimize_scalar

# relative
from ...core.node.common.node_manager.user_manager import RefreshBudgetException
from ...lib.numpy.array import capnp_deserialize
from ...lib.numpy.array import capnp_serialize
from ..common.serde.capnp import CapnpModule
from ..common.serde.capnp import get_capnp_schema
from ..common.serde.capnp import serde_magic_header
from ..common.serde.serializable import serializable
from .abstract_ledger_store import AbstractDataSubjectLedger
from .abstract_ledger_store import AbstractLedgerStore
from .data_subject_list import DataSubjectArray


def convert_constants_to_indices(rdp_constant_array: np.ndarray) -> np.ndarray:
    """
    Given an array of RDP Constants, this will return an array of the same size/shape telling you which indices in the
    DataSubjectLedger's cache you need to query.

    This currently assumes the cache generated on May 4th 2022, where there are 1.2M values in total.
    - 500,000 of these correspond to RDP constants between 0 and 50 (10,000 between any two consecutive integers)
    - 700,000 of these correspond to RDP constants between 50 and 700,050

    An easy way to check if you're using the right cache is that the very
    first value in the cache should be 0.05372712063485988

    MAKE SURE THERE ARE NO ZEROS IN THE CACHE!!
    """
    # Find indices for all RDP constants <= 50
    sub50_mask = rdp_constant_array <= 50
    sub50_indices = (((rdp_constant_array - 1) * sub50_mask) * 10_000).astype(int)

    # Find indices for all RDP constants > 50
    gt50_mask = rdp_constant_array > 50
    gt50_indices = ((rdp_constant_array - 51 + 500_000) * gt50_mask).astype(int)

    # We should be able to do a straight addition because
    return sub50_indices + gt50_indices


def get_cache_path(cache_filename: str) -> str:
    here = os.path.dirname(__file__)
    root_dir = Path(here) / ".." / ".." / "cache"
    return os.path.abspath(root_dir / cache_filename)


def load_cache(filename: str) -> np.ndarray:
    CACHE_PATH = get_cache_path(filename)
    if not os.path.exists(CACHE_PATH):
        raise Exception(f"Cannot load {CACHE_PATH}")
    cache_array = np.load(CACHE_PATH)
    info(f"Loaded constant2epsilon cache of size: {cache_array.shape}")
    return cache_array


@dataclass
class RDPParams:
    sigmas: jnp.array
    l2_norms: jnp.array
    l2_norm_bounds: jnp.array
    Ls: jnp.array
    coeffs: jnp.array

    def __repr__(self) -> str:
        res = "RDPParams:"
        res = f"{res}\n sigmas:{self.sigmas}"
        res = f"{res}\n l2_norms:{self.l2_norms}"
        res = f"{res}\n l2_norm_bounds:{self.l2_norm_bounds}"
        res = f"{res}\n Ls:{self.Ls}"
        res = f"{res}\n coeffs:{self.coeffs}"

        return res


def get_unique_data_subjects(data_subjects_query: np.ndarray) -> np.ndarray:
    # This might look horribly wrong, but .sum() returns all the unique DS ~ Ishan
    return sorted(list(data_subjects_query.sum()))


def convert_dsa_to_index_array(
    data_subject_array: np.ndarray,
) -> Tuple[np.ndarray, int]:
    """Convert data subject array to data subject index array."""

    unique_data_subjects = get_unique_data_subjects(data_subject_array)
    max_entity = len(unique_data_subjects)

    input_entities_indexes_list: List[np.ndarray] = []

    for data_subject_idx, data_subject in enumerate(unique_data_subjects):
        # Create a mask where the current data subject is present
        data_subject = DataSubjectArray([data_subject])
        ds_mask = np.isin(data_subject_array, data_subject)
        input_entity_indexes = ds_mask * (
            np.ones_like(data_subject_array, np.int64) * (data_subject_idx + 1)
        )
        input_entities_indexes_list.append(input_entity_indexes)

    input_entities_indexes: np.ndarray = np.stack(input_entities_indexes_list)

    return input_entities_indexes, max_entity


# @partial(jax.jit, static_argnums=3, donate_argnums=(1, 2))
def first_try_branch(
    constant: jax.numpy.DeviceArray,
    rdp_constants: np.ndarray,
    entity_ids_query: np.ndarray,
) -> jax.numpy.DeviceArray:

    input_entities_indexes, max_entity = convert_dsa_to_index_array(entity_ids_query)

    if max_entity < len(rdp_constants):
        # Take only the constants values where current data subject is present
        summed_constant = constant.take(input_entities_indexes) + rdp_constants.take(
            input_entities_indexes
        )

        # Set rpd constants for given data subjects
        rdp_constants[input_entities_indexes] = summed_constant
    else:
        pad_length = max_entity - len(rdp_constants) + 1
        rdp_constants = jnp.concatenate([rdp_constants, jnp.zeros(shape=pad_length)])

        # Take only the constants values where current data subject is present
        summed_constant = constant.take(input_entities_indexes) + rdp_constants.take(
            input_entities_indexes
        )

        # Set rpd constants for given data subjects
        # jax.interpreters.xla._DeviceArray does not support item assignment
        rdp_constants = rdp_constants.at[input_entities_indexes].set(summed_constant)

    return rdp_constants


@partial(jax.jit, static_argnums=1)
def compute_rdp_constant(rdp_params: RDPParams, private: bool) -> jax.numpy.DeviceArray:
    squared_Ls = rdp_params.Ls**2
    squared_sigma = rdp_params.sigmas**2

    if private:
        # this is calculated on the private true values
        squared_l2 = rdp_params.l2_norms**2
    else:
        # bounds is computed on the metadata
        squared_l2 = rdp_params.l2_norm_bounds**2

    return (squared_Ls * squared_l2 / (2 * squared_sigma)) * rdp_params.coeffs


@jax.jit
def get_budgets_and_mask(
    epsilon_spend: jnp.array, user_budget: jnp.float64
) -> Tuple[float, float, jax.numpy.DeviceArray]:
    # Function to vectorize the result of the budget computation.
    mask = jnp.ones_like(epsilon_spend) * user_budget < epsilon_spend
    # get the highest value which was under budget and represented by False in the mask
    highest_possible_spend = jnp.max(epsilon_spend * (1 - mask))
    return (highest_possible_spend, user_budget, mask)


@serializable(capnp_bytes=True)
class DataSubjectLedger(AbstractDataSubjectLedger):
    """for a particular data subject, this is the list
    of all mechanisms releasing informationo about this
    particular subject, stored in a vectorized form"""

    CONSTANT2EPSILSON_CACHE_FILENAME = "constant2epsilon_1200k.npy"
    _cache_constant2epsilon = load_cache(filename=CONSTANT2EPSILSON_CACHE_FILENAME)

    def __init__(
        self,
        constants: Optional[np.ndarray] = None,
        update_number: int = 0,
        timestamp_of_last_update: Optional[float] = None,
    ) -> None:
        self._rdp_constants = (
            constants if constants is not None else np.array([], dtype=np.float64)
        )
        self._update_number = update_number
        self._timestamp_of_last_update = (
            timestamp_of_last_update
            if timestamp_of_last_update is not None
            else time.time()
        )
        self._pending_save = False

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, DataSubjectLedger):
            return self == other
        return (
            self._update_number == other._update_number
            and self._timestamp_of_last_update == other._timestamp_of_last_update
            and all(self._rdp_constants == other._rdp_constants)
        )

    @property
    def delta(self) -> float:
        FIXED_DELTA: Final = 1e-6
        return FIXED_DELTA  # WARNING: CHANGING DELTA INVALIDATES THE CACHE

    def bind_to_store_with_key(
        self, store: AbstractLedgerStore, user_key: VerifyKey
    ) -> None:
        self.store = store
        self.user_key = user_key

    @staticmethod
    def get_or_create(
        store: AbstractLedgerStore, user_key: VerifyKey
    ) -> Optional[AbstractDataSubjectLedger]:
        ledger: Optional[AbstractDataSubjectLedger] = None
        try:
            # todo change user_key or uid?
            ledger = store.get(key=user_key)
            ledger.bind_to_store_with_key(store=store, user_key=user_key)
        except KeyError:
            print("Creating new Ledger")
            ledger = DataSubjectLedger()
            ledger.bind_to_store_with_key(store=store, user_key=user_key)
        except Exception as e:
            print(f"Failed to read ledger from ledger store. {e}")

        return ledger

    def get_entity_overbudget_mask_for_epsilon_and_append(
        self,
        unique_entity_ids_query: np.ndarray,
        rdp_params: RDPParams,
        get_budget_for_user: Callable,
        deduct_epsilon_for_user: Callable,
        private: bool = True,
    ) -> np.ndarray:
        # coerce to np.int64
        entity_ids_query: np.ndarray = (
            unique_entity_ids_query  # = unique_entity_ids_query.astype(np.int64)
        )
        # calculate constants
        rdp_constants = self._get_batch_rdp_constants(
            entity_ids_query=entity_ids_query, rdp_params=rdp_params, private=private
        )
        print("rdp constants", rdp_constants)

        # here we iteratively attempt to calculate the overbudget mask and save
        # changes to the database
        mask = self._get_overbudgeted_entities(
            get_budget_for_user=get_budget_for_user,
            deduct_epsilon_for_user=deduct_epsilon_for_user,
            rdp_constants=rdp_constants,
        )

        # at this point we are confident that the database budget field has been updated
        # so now we should flush the _rdp_constants that we have calculated to storage
        if self._write_ledger():
            return mask

    def _write_ledger(self) -> bool:

        self._update_number += 1
        try:
            self._pending_save = False
            self.store.set(key=self.user_key, value=self)
            return True
        except Exception as e:
            self._pending_save = True
            print(f"Failed to write ledger to ledger store. {e}")
            raise e

    def _increase_max_cache(self, new_size: int) -> None:
        new_entries = []
        current_size = len(self._cache_constant2epsilon)
        new_alphas = []
        for i in range(new_size - current_size):
            alph, eps = self._get_optimal_alpha_for_constant(
                constant=i + 1 + current_size
            )
            new_entries.append(eps)
            new_alphas.append(alph)

        self._cache_constant2epsilon = np.concatenate(
            [self._cache_constant2epsilon, np.array(new_entries)]
        )

    def _get_fake_rdp_func(self, constant: int) -> Callable:
        def func(alpha: float) -> float:
            return alpha * constant

        return func

    def _get_alpha_search_function(self, rdp_compose_func: Callable) -> Callable:
        log_delta = np.log(self.delta)

        def fun(alpha: float) -> float:  # the input is the RDP's \alpha
            if alpha <= 1:
                return np.inf
            else:
                alpha_minus_1 = alpha - 1
                return np.maximum(
                    rdp_compose_func(alpha)
                    + np.log(alpha_minus_1 / alpha)
                    - (log_delta + np.log(alpha)) / alpha_minus_1,
                    0,
                )

        return fun

    def _get_optimal_alpha_for_constant(
        self, constant: int = 3
    ) -> Tuple[np.ndarray, Callable]:
        f = self._get_fake_rdp_func(constant=constant)
        f2 = self._get_alpha_search_function(rdp_compose_func=f)
        results = minimize_scalar(
            f2, method="Brent", bracket=(1, 2), bounds=[1, np.inf]
        )

        return results.x, results.fun

    def _get_batch_rdp_constants(
        self, entity_ids_query: jnp.ndarray, rdp_params: RDPParams, private: bool = True
    ) -> jnp.ndarray:
        constant = compute_rdp_constant(rdp_params, private)
        if self._rdp_constants.size == 0:
            self._rdp_constants = np.zeros_like(np.asarray(constant, constant.dtype))
        # print("constant: ", constant)
        # print("_rdp_constants: ", self._rdp_constants)
        # print("entity ids query", entity_ids_query)
        # print(jnp.max(entity_ids_query))
        self._rdp_constants = first_try_branch(
            constant, self._rdp_constants, entity_ids_query
        )
        return constant

    def _get_epsilon_spend(self, rdp_constants: np.ndarray) -> np.ndarray:
        # rdp_constants_lookup = (rdp_constants - 1).astype(np.int64)
        rdp_constants_lookup = convert_constants_to_indices(rdp_constants)
        try:
            # needed as np.int64 to use take
            eps_spend = jax.jit(jnp.take)(
                self._cache_constant2epsilon, rdp_constants_lookup
            )
        except IndexError:
            print(f"Cache missed the value at {max(rdp_constants_lookup)}")
            self._increase_max_cache(int(max(rdp_constants_lookup) * 1.1))
            eps_spend = jax.jit(jnp.take)(
                self._cache_constant2epsilon, rdp_constants_lookup
            )
        return eps_spend

    def _calculate_mask_for_current_budget(
        self, get_budget_for_user: Callable, epsilon_spend: np.ndarray
    ) -> Tuple[float, float, np.ndarray]:
        user_budget = get_budget_for_user(verify_key=self.user_key)
        # create a mask of True and False where true is over current user_budget
        return get_budgets_and_mask(epsilon_spend, user_budget)

    def _get_overbudgeted_entities(
        self,
        get_budget_for_user: Callable,
        deduct_epsilon_for_user: Callable,
        rdp_constants: np.ndarray,
    ) -> Tuple[np.ndarray]:
        """TODO:
        In our current implementation, user_budget is obtained by querying the
        Adversarial Accountant's entity2ledger with the Data Scientist's User Key.
        When we replace the entity2ledger with something else, we could perhaps directly
        add it into this method
        """
        epsilon_spend = self._get_epsilon_spend(rdp_constants=rdp_constants)

        # try first time
        (
            highest_possible_spend,
            user_budget,
            mask,
        ) = self._calculate_mask_for_current_budget(
            get_budget_for_user=get_budget_for_user, epsilon_spend=epsilon_spend
        )

        mask = np.array(mask, copy=False)
        highest_possible_spend = float(highest_possible_spend)
        user_budget = float(user_budget)
        print("Epsilon spend ", epsilon_spend)
        print("Highest possible spend ", highest_possible_spend)
        if highest_possible_spend > 0:
            # go spend it in the db
            attempts = 0
            while attempts < 5:
                print(
                    f"Attemping to spend epsilon: {highest_possible_spend}. Try: {attempts}"
                )
                attempts += 1
                try:
                    user_budget = self.spend_epsilon(
                        deduct_epsilon_for_user=deduct_epsilon_for_user,
                        epsilon_spend=highest_possible_spend,
                        old_user_budget=user_budget,
                    )
                    break
                except RefreshBudgetException:  # nosec
                    # this is the only exception we allow to retry
                    (
                        highest_possible_spend,
                        user_budget,
                        mask,
                    ) = self._calculate_mask_for_current_budget(
                        get_budget_for_user=get_budget_for_user,
                        epsilon_spend=epsilon_spend,
                    )
                except Exception as e:
                    print(f"Problem spending epsilon. {e}")
                    raise e

        if user_budget is None:
            raise Exception("Failed to spend_epsilon")

        return mask

    def spend_epsilon(
        self,
        deduct_epsilon_for_user: Callable,
        epsilon_spend: float,
        old_user_budget: float,
    ) -> float:
        # get the budget
        print("got user budget", old_user_budget, "epsilon_spent", epsilon_spend)
        deduct_epsilon_for_user(
            verify_key=self.user_key,
            old_budget=old_user_budget,
            epsilon_spend=epsilon_spend,
        )
        # return the budget we used
        return old_user_budget

    def _object2bytes(self) -> bytes:
        schema = get_capnp_schema(schema_file="data_subject_ledger.capnp")

        dsl_struct: CapnpModule = schema.DataSubjectLedger  # type: ignore
        dsl_msg = dsl_struct.new_message()
        # this is how we dispatch correct deserialization of bytes
        dsl_msg.magicHeader = serde_magic_header(type(self))
        self._rdp_constants = np.array(self._rdp_constants, copy=False)

        dsl_msg.constants = capnp_serialize(self._rdp_constants)
        dsl_msg.updateNumber = self._update_number
        dsl_msg.timestamp = self._timestamp_of_last_update

        return dsl_msg.to_bytes_packed()

    @staticmethod
    def _bytes2object(buf: bytes) -> DataSubjectLedger:
        schema = get_capnp_schema(schema_file="data_subject_ledger.capnp")
        dsl_struct: CapnpModule = schema.DataSubjectLedger  # type: ignore
        # https://stackoverflow.com/questions/48458839/capnproto-maximum-filesize
        MAX_TRAVERSAL_LIMIT = 2**64 - 1
        # to pack or not to pack?
        dsl_msg = dsl_struct.from_bytes_packed(
            buf, traversal_limit_in_words=MAX_TRAVERSAL_LIMIT
        )

        constants = capnp_deserialize(dsl_msg.constants)
        update_number = dsl_msg.updateNumber
        timestamp_of_last_update = dsl_msg.timestamp

        return DataSubjectLedger(
            constants=constants,
            update_number=update_number,
            timestamp_of_last_update=timestamp_of_last_update,
        )



# stdlib
from typing import List
from typing import Optional

# third party
from nacl.signing import VerifyKey

# relative
from ......logger import debug
from ......logger import error
from ......logger import traceback_and_raise
from .....common.message import ImmediateSyftMessageWithReply
from .....common.message import ImmediateSyftMessageWithoutReply
from .....common.message import SignedImmediateSyftMessageWithReply
from .....common.message import SignedImmediateSyftMessageWithoutReply
from .....common.message import SignedMessageT
from ....abstract.node import AbstractNode
from ...client import GET_OBJECT_TIMEOUT
from ..node_service import SignedNodeServiceWithReply
from ..node_service import SignedNodeServiceWithoutReply


class SignedMessageWithoutReplyForwardingService(SignedNodeServiceWithoutReply):
    @staticmethod
    def process(
        node: AbstractNode,
        msg: SignedImmediateSyftMessageWithoutReply,
        verify_key: Optional[VerifyKey] = None,
    ) -> Optional[SignedMessageT]:
        addr = msg.address
        debug(f"> Forwarding WithoutReply {msg.pprint} to {addr.target_emoji()}")
        # order is important, vm, device, domain, network
        for scope_id in [addr.vm_id, addr.device_id, addr.domain_id, addr.network_id]:
            if scope_id is not None and scope_id in node.store:
                obj = node.store.get(scope_id)
                func = getattr(obj, "send_immediate_msg_without_reply", None)

                if func is None:
                    error(
                        f"{addr} in store does not have method send_immediate_msg_without_reply"
                    )
                else:
                    return func(msg=msg)

        try:
            for scope_id in [
                addr.vm_id,
                addr.device_id,
                addr.domain_id,
                addr.network_id,
            ]:
                if scope_id is not None:
                    debug(f"> Lookup: {scope_id.emoji()}")
                    client = node.get_peer_client(node_id=scope_id, only_vpn=False)
                    if client:
                        return client.send_immediate_msg_without_reply(msg=msg)
                    else:
                        raise Exception
        except Exception as e:
            # TODO: Need to not catch blanket exceptions
            error(
                f"Failed to forward {msg}. "
                f"No client for {addr} in node.get_peer_client. {e}"
            )
            pass
        debug(f"> ❌ {node.pprint} 🤷🏾‍♀️ {addr.target_emoji()}")
        traceback_and_raise(
            Exception("Address unknown - cannot forward message. Throwing it away.")
        )

    @staticmethod
    def message_handler_types() -> List[type]:
        return [ImmediateSyftMessageWithoutReply]


class SignedMessageWithReplyForwardingService(SignedNodeServiceWithReply):
    @staticmethod
    def process(
        node: AbstractNode,
        msg: SignedImmediateSyftMessageWithReply,
        verify_key: Optional[VerifyKey] = None,
    ) -> SignedImmediateSyftMessageWithoutReply:
        addr = msg.address
        debug(f"> Forwarding WithReply {msg.pprint} to {addr.target_emoji()}")

        # order is important, vm, device, domain, network
        for scope_id in [addr.vm_id, addr.device_id, addr.domain_id, addr.network_id]:
            if scope_id is not None and scope_id in node.store:
                obj = node.store.get(scope_id)
                func = getattr(obj, "send_immediate_msg_with_reply", None)
                if func is None or not callable(func):
                    error(
                        f"{addr} in store does not have method send_immediate_msg_with_reply"
                    )
                else:
                    return func(msg=msg, timeout=15)

        try:
            for scope_id in [
                addr.vm_id,
                addr.device_id,
                addr.domain_id,
                addr.network_id,
            ]:
                if scope_id is not None:
                    debug(f"> Lookup: {scope_id.emoji()}")
                    client = node.get_peer_client(node_id=scope_id, only_vpn=False)
                    if client:
                        return client.send_immediate_msg_with_reply(msg=msg, timeout=GET_OBJECT_TIMEOUT)  # type: ignore
                    else:
                        raise Exception
        except Exception as e:
            # TODO: Need to not catch blanket exceptions
            error(
                f"Failed to forward {msg}. "
                f"No client for {addr} in node.get_peer_client. {e}"
            )
            pass
        debug(f"> ❌ {node.pprint} 🤷🏾‍♀️ {addr.target_emoji()}")
        traceback_and_raise(
            Exception("Address unknown - cannot forward message. Throwing it away.")
        )

    @staticmethod
    def message_handler_types() -> List[type]:
        return [ImmediateSyftMessageWithReply]
